unit ASpellChecker;

interface

uses
  SpellChecker,
  AspellHeadersDyn;

type
  TASpellChecker = class(TCustomSpellChecker)
  private
    FSpellChecker: AspellSpeller;
  protected
    procedure LoadDictionaries; override;
    procedure RealizeDictionary; override;
    procedure CloseDictionary;
  public
    constructor Create; override;
    destructor Destroy; override;

    function SpelledCorrect(const AWord: WideString): Boolean; override;
    procedure FillSuggestions(const AWord: WideString); override;
    procedure Replace(const AWord, AReplacer: WideString); override;
    procedure Add(const AWord: WideString); override;
    procedure Ignore(const AWord: WideString); override;
  end;

implementation

uses
  SysUtils, MWKEdit;

{ TASpellChecker }

procedure TASpellChecker.Add(const AWord: WideString);
begin
  ASpell_Speller_Add_To_PersonalW(FSpellChecker, PWideChar(AWord), -1);
end;

procedure TASpellChecker.CloseDictionary;
begin
  if Assigned(FSpellChecker) then
  begin
    aspell_speller_save_all_word_lists(FSpellChecker);
    delete_aspell_speller(FSpellChecker);
    FSpellChecker := nil;
  end;
end;

constructor TASpellChecker.Create;
begin
  inherited;
  FLoaded := LoadAspell('');
  LoadDictionaries;
end;

destructor TASpellChecker.Destroy;
begin
  CloseDictionary;
  inherited;
end;

procedure TASpellChecker.FillSuggestions(const AWord: WideString);
var
  Elements: AspellStringEnumeration;
  Element: PWideChar;
  WordList: AspellWordList;
begin
  Suggestions.Clear;
  WordList := ASpell_Speller_SuggestW(FSpellChecker, PWideChar(AWord), -1);
  Elements := aspell_word_list_elements(WordList);

  repeat
    Element := ASpell_String_Enumeration_NextW(Elements);
    if Assigned(Element) then
      Suggestions.Add(Element);
  until (Element = nil);

  delete_aspell_string_enumeration(Elements);
end;

procedure TASpellChecker.Ignore(const AWord: WideString);
begin
  ASpell_Speller_Add_To_SessionW(FSpellChecker, PWideChar(AWord), -1);
end;

procedure TASpellChecker.LoadDictionaries;
var
  cnf: AspellConfig;
  dict_info_list: AspellDictInfoList;
  dict_info_elements: AspellDictInfoEnumeration;
  dict_info: TAspellDictInfo;
begin
  Dictionaries.Clear;

  if not FLoaded then
    Exit;
  cnf := new_aspell_config();
  dict_info_list := get_aspell_dict_info_list(cnf);
  delete_aspell_config(cnf);
  dict_info_elements := aspell_dict_info_list_elements(dict_info_list);
  repeat
    dict_info := aspell_dict_info_enumeration_next(dict_info_elements)^;
    Dictionaries.Add(dict_info.name);
  until (aspell_dict_info_enumeration_at_end(dict_info_elements) <> 0);
  delete_aspell_dict_info_enumeration(dict_info_elements);
end;

procedure TASpellChecker.RealizeDictionary;
var
  cnf: AspellConfig;
  possible_err: AspellCanHaveError;
begin
  CloseDictionary;

  cnf := new_aspell_config();
  aspell_config_replace(cnf, 'lang', PChar(AnsiString(Dictionaries[CurrentDictionary])));
  aspell_config_replace(cnf, 'encoding', 'machine unsigned 16');

  possible_err := new_aspell_speller(cnf);
  if (aspell_error_number(possible_err) <> 0) then
  begin
    delete_aspell_config(cnf);
    Exit;
  end;

  FSpellChecker := to_aspell_speller(possible_err);
  delete_aspell_config(cnf);
end;

procedure TASpellChecker.Replace(const AWord, AReplacer: WideString);
begin
  ASpell_Speller_Store_ReplacementW(FSpellChecker, PWideChar(AWord), -1,
    PWideChar(AReplacer), -1);
end;

function TASpellChecker.SpelledCorrect(const AWord: WideString): Boolean;
begin
  Result := ASpell_Speller_CheckW(FSpellChecker, PWideChar(AWord), -1) = 1;
end;

initialization
  TMWKEdit.RegisterSpellChecker('ASpell', TASpellChecker);
end.
