unit HunSpellChecker;

interface

uses
  SpellChecker,
  WideStringList,
  uHunSpellLib;

type
  TDictionaryEncoding = (deANSI, deUTF8, deUTF16, deWide);

  THunSpellChecker = class(TCustomSpellChecker)
  private
    FSpell: Pointer;
    FUserDictionaryDir: WideString;
    FUserDictionary: TWideStringList;
    FUserFileName: WideString;
    FUpdateUserFile: Boolean;
    FEncoding: TDictionaryEncoding;

    function Encode(const AWord: WideString): AnsiString;
    function Decode(const AWord: AnsiString): WideString;
  protected
    procedure LoadDictionaries; override;
    function AddToUserDir(const AWord: WideString): Boolean;
    procedure RealizeDictionary; override;
    procedure UpdateUserDirectory;
    procedure CloseDictionary;
  public
    constructor Create; override;
    destructor Destroy; override;

    function SpelledCorrect(const AWord: WideString): Boolean; override;
    procedure FillSuggestions(const AWord: WideString); override;
    procedure Replace(const AWord, AReplacer: WideString); override;
    procedure Add(const AWord: WideString); override;
    procedure Ignore(const AWord: WideString); override;
  end;

implementation

uses
  Dialogs, ShlObj, SysUtils,
  MWKSysUtils, MWKEdit;

{ THunSpellChecker }

procedure THunSpellChecker.Add(const AWord: WideString);
begin
  if AddToUserDir(AWord) then
    FUpdateUserFile := True;
end;

function THunSpellChecker.AddToUserDir(const AWord: WideString): Boolean;
var
  S: WideString;
  C: Integer;
begin
  Result := False;
  S := WideLowerCase(AWord);

  for C := 0 to FUserDictionary.Count - 1 do
    if WideLowerCase(FUserDictionary[C]) = S then
      Exit;

  FUserDictionary.Add(AWord);
  Result := True;
end;

procedure THunSpellChecker.CloseDictionary;
begin
  if Assigned(FSpell) then
  begin
    UpdateUserDirectory;
    hunspell_uninitialize(FSpell);
    FSpell := nil;
  end;
end;

constructor THunSpellChecker.Create;
begin
  inherited;

  FUserDictionaryDir := AddTrailingPathDelimiter(GetShellFolder(CSIDL_APPDATA));
  FUserDictionary := TWideStringList.Create;
  FUserDictionary.AllowDuplicates := False;

  FLoaded := LoadLibHunspell('');
  LoadDictionaries;
end;

function THunSpellChecker.Decode(const AWord: AnsiString): WideString;
begin
  case FEncoding of
    deANSI:
      Result := AWord;
    deUTF8:
      Result := UTF8Decode(AWord);
    deUTF16:
      Result := AWord;
    deWide:
      Result := AWord;
  end;
end;

destructor THunSpellChecker.Destroy;
begin
  CloseDictionary;

  FUserDictionary.Free;

  inherited;
end;

function THunSpellChecker.Encode(const AWord: WideString): AnsiString;
begin
  case FEncoding of
    deANSI:
      Result := AWord;
    deUTF8:
      Result := UTF8Encode(AWord);
    deUTF16:
      Result := AWord;
    deWide:
      Result := AWord;
  end;
end;

procedure THunSpellChecker.FillSuggestions(const AWord: WideString);
var
  Words, Wrds: PPChar;
  C, L: Integer;
begin
  Suggestions.Clear;
  if not FLoaded then
    Exit;
  L := hunspell_suggest(FSpell, PChar(Encode(AWord)), Words);

  Wrds := Words;
  for C := 1 to L do
  begin
    Suggestions.Add(Decode(Words^));
    Inc(Integer(Words), SizeOf(Pointer));
  end;

  hunspell_suggest_free(FSpell, Wrds, L);
end;

procedure THunSpellChecker.Ignore(const AWord: WideString);
begin
  hunspell_put_word(FSpell, PChar(Encode(AWord)));
end;

procedure THunSpellChecker.LoadDictionaries;
var
  SR: TSearchRec;
begin
  Dictionaries.Clear;
  if FindFirst(DictionaryDir + '*.aff', faAnyFile, SR) = 0 then
  begin
    Dictionaries.Add(ChangeFileExt(SR.Name, ''));
    while FindNext(SR) = 0 do
      Dictionaries.Add(ChangeFileExt(SR.Name, ''));
  end;
  FindClose(SR);
end;

procedure THunSpellChecker.RealizeDictionary;
var
  UserWord: AnsiString;
  Language: WideString;
  Encoding: PChar;
  C: Integer;
begin
  CloseDictionary;

  if not Loaded then
    Exit;

  Language := Dictionaries[CurrentDictionary];

  FSpell := hunspell_initialize(PChar(AnsiString(DictionaryDir + Language + '.aff')),
    PChar(AnsiString(DictionaryDir + Language + '.dic')));

  Encoding := hunspell_get_dic_encoding(FSpell);

  FEncoding := deANSI;
  if (Encoding = 'UTF-8') then
    FEncoding := deUTF8;

  FUserFileName := FUserDictionaryDir + Language + '.user.dic';

  FUserDictionary.Clear;
  if FileExists(FUserFileName) then
  begin
    FUserDictionary.LoadFromFile(FUserFileName);
    for C := 0 to FUserDictionary.Count - 1 do
    begin
      UserWord := Encode(Trim(FUserDictionary[C]));
      if UserWord <> '' then
        hunspell_put_word(FSpell, PChar(UserWord));
    end;
  end;
  FUpdateUserFile := False;
end;

procedure THunSpellChecker.Replace(const AWord, AReplacer: WideString);
begin
  // Nothing to do here
end;

function THunSpellChecker.SpelledCorrect(const AWord: WideString): Boolean;
begin
  if FLoaded then
    Result := hunspell_spell(FSpell, PChar(Encode(AWord)))
  else
    Result := True;
end;

procedure THunSpellChecker.UpdateUserDirectory;
var
  C: Integer;
  SL: TWideStringList;
begin
  if FUpdateUserFile then
  begin
    if FileExists(FUserFileName) then
    begin
      SL := TWideStringList.Create;
      try
        SL.LoadFromFile(FUserFileName);
        for C := 0 to SL.Count - 1 do
          AddToUserDir(SL[C]);
      finally
        SL.Free;
      end;
    end;

    FUserDictionary.SaveToFile(FUserFileName);

    FUpdateUserFile := False;
  end;
end;

initialization
  TMWKEdit.RegisterSpellChecker('HunSpell', THunSpellChecker);
end.
