unit MWKEditGraphics;
      {******************************************************************}
      { MWKEditGraphics                                                  }
      {                                                                  }
      { home page : http://www.mwcs.de                                   }
      { email     : martin.walter@mwcs.de                                }
      {                                                                  }
      { date      : 27-09-2006                                           }
      { version   : v1.0                                                 }
      {                                                                  }
      { Use of this file is permitted for commercial and non-commercial  }
      { use, as long as the original author is credited.                 }
      { This file (c) 2005, 2006 Martin Walter                           }
      {                                                                  }
      { This Software is distributed on an "AS IS" basis, WITHOUT        }
      { WARRANTY OF ANY KIND, either express or implied.                 }
      {                                                                  }
      { *****************************************************************}

interface

uses
  Graphics,         
  MSXML2_TLB_Light,
  GDIPOBJ, GDIPAPI;

function CalcImageWidth(FImage: Pointer; Height: Integer): Integer;

procedure PaintImage(FImage: Pointer; Canvas: TCanvas;
  X, Y, Width, Height: Integer; Selected, Focused: Boolean);

function CloneImage(FImage: Pointer): Pointer;

function ImageToText(FImage: Pointer; WithMarker: Boolean): WideString;

function TextToImage(Marker: WideChar; const Text: WideString): TObject;

function IsImageSVG(Image: TObject): Boolean;

function GetSVGXML(Image: TObject): IXMLDOMDocument3;

function LoadImage(const FileName: WideString): TObject;

implementation

uses
  Classes, ActiveX, SysUtils,
  GDIPUtil, SVG, MWKSysUtils;

function CalcImageWidth(FImage: Pointer; Height: Integer): Integer;
var
  W, H, D: Double;
  SVG: TSVG;
  Bitmap: TGPImage;
begin
  if FImage = nil then
  begin
    Result := 0;
    Exit;
  end;

  if TObject(FImage) is TSVG  then
  begin
    SVG := TSVG(FImage);
    W := SVG.Width;
    H := SVG.Height;
  end else
  begin
    Bitmap := TGPImage(FImage);
    W := Bitmap.GetWidth;
    H := Bitmap.GetHeight;
  end;

  if H > 0 then
    D :=  W / H
  else
    D := 1;

  if (Height <> 0)  then
    Result := Round(Height * D) + 1
  else
    Result := 0;
end;

function GraphicRect(W, H: Double; R: TGPRectF): TGPRectF;
var
  D: Double;
  Width, Height: Double;
begin
  Width := R.Width;
  Height := R.Height;

  if H > 0 then
    D :=  W / H
  else
    D := 1;

  if (Height <> 0) and
     (Width / Height > D) then
  begin
    Result.Width := Height * D;
    Result.Height := Height;
  end else
  begin
    Result.Width := Width;
    Result.Height := Width / D;
  end;

  Result.X := R.X;
  Result.Y := R.Y;
end;

procedure PaintImage(FImage: Pointer; Canvas: TCanvas;
  X, Y, Width, Height: Integer; Selected, Focused: Boolean);
var
  Rect: TGPRectF;
  SVG: TSVG;
  Bitmap: TGPImage;
  Graphics: TGPGraphics;
  R: TGPRectF;
  Brush: TGPBrush;
begin
  if FImage = nil then
    Exit;
  Graphics := TGPGraphics.Create(Canvas.Handle);

  R.X := X;
  R.Y := Y;
  R.Width := Width - 1;
  R.Height := Height;

  if TObject(FImage) is TSVG then
  begin
    SVG := TSVG(FImage);
    Rect := GraphicRect(SVG.Width, SVG.Height, R);
    Rect.Width := Rect.Width - 1;
    SVG.PaintTo(Graphics, Rect, nil, 0);
  end else
  begin
    Bitmap := TGPImage(FImage);
    Rect := GraphicRect(Bitmap.GetWidth, Bitmap.GetHeight, R);

    Graphics.DrawImage(Bitmap, Rect);
  end;

  if Selected then
  begin
    if Focused then
      Brush := TGPSolidBrush.Create(MakeColor(128, 128, 128, 128))
    else
      Brush := TGPSolidBrush.Create(MakeColor(192, 128, 128, 128));

    Graphics.FillRectangle(Brush, Rect);
    Brush.Free;
  end;
  Graphics.Free;
end;

function CloneImage(FImage: Pointer): Pointer;
var
  SVG: TSVG;
  Bitmap: TGPImage;
begin
  if FImage = nil then
  begin
    Result := nil;
    Exit;
  end;

  if TObject(FImage) is TSVG then
  begin
    SVG := TSVG(FImage);
    Result := SVG.Clone(nil);
  end else
  begin
    Bitmap := TGPImage(FImage);
    Result := Bitmap.Clone;
  end;
end;

function ImageToText(FImage: Pointer; WithMarker: Boolean): WideString;
var
  SVG: TSVG;
  Bitmap: TGPImage;
  EncoderID: TGUID;
  St: TMemoryStream;
  SA: IStream;
begin
  if FImage = nil then
  begin
    Result := '';
    Exit;
  end;

  if TObject(FImage) is TSVG then
  begin
    SVG := TSVG(FImage);
    if WithMarker then
      Result := '1' + SVG.Source
    else
      Result := SVG.Source;
  end else
  begin
    Bitmap := TGPImage(FImage);
    if GetEncoderClsid('image/png', EncoderID) <> -1 then
    begin
      St := TMemoryStream.Create;
      SA := TStreamAdapter.Create(St);
      Bitmap.Save(SA, EncoderID);
      if WithMarker then
        Result := '2' + EncodeBase64(St)
      else
        Result := EncodeBase64(St);
      SA := nil;
      St.Free;
    end;
  end;
end;

function TextToImage(Marker: WideChar; const Text: WideString): TObject;
var
  SVG: TSVG;
  Bitmap: TGPImage;
  St: TMemoryStream;
  SA: IStream;
begin
  Result := nil;
  if Marker = '1' then
  begin
    SVG := TSVG.Create;
    SVG.LoadFromText(Text);
    Result := SVG;
  end;

  if Marker = '2' then
  begin
    St := TMemoryStream.Create;
    DecodeBase64(Text, St);
    St.Position := 0;
    SA := TStreamAdapter.Create(St);
    Bitmap := TGPImage.Create(SA);
    St.Free;
    Result := Bitmap;
  end;
end;

function IsImageSVG(Image: TObject): Boolean;
begin
  Result := Image is TSVG;
end;

function GetSVGXML(Image: TObject): IXMLDOMDocument3;
begin
  if Image = nil then
    Exit;

  Result := CoDOMDocument60.Create;
  Result.setProperty('NewParser', True);
  Result.setProperty('ProhibitDTD', False);
  Result.validateOnParse := False;
  Result.resolveExternals := False;
  Result.loadXML(TSVG(Image).Source);
end;

function LoadImage(const FileName: WideString): TObject;
var
  SVG: TSVG;
  Image: TGPImage;
begin
  Result := nil;
  if LowerCase(ExtractFileExt(FileName)) = '.svg' then
  begin
    SVG := TSVG.Create;
    try
      SVG.LoadFromFile(FileName);
      if SVG.Source <> '' then
        Result := SVG;
    except
      SVG.Free;
    end
  end else
  begin
    Image := TGPImage.Create(FileName);
    if Image.GetLastStatus = Ok then
      Result := Image
    else
      Image.Free;
  end;
end;

end.