unit MWKEditRTF2XML;
      {******************************************************************}
      { MWKEditRTF2XML                                                   }
      {                                                                  }
      { home page : http://www.mwcs.de                                   }
      { email     : martin.walter@mwcs.de                                }
      {                                                                  }
      { date      : 02-12-2007                                           }
      { version   : v1.3                                                 }
      {                                                                  }
      { Use of this file is permitted for commercial and non-commercial  }
      { use, as long as the original author is credited.                 }
      { This file (c) 2005, 2007 Martin Walter                           }
      {                                                                  }
      { This Software is distributed on an "AS IS" basis, WITHOUT        }
      { WARRANTY OF ANY KIND, either express or implied.                 }
      {                                                                  }
      { *****************************************************************}

interface

uses
  Classes, Contnrs, MWKEditSpans,
  MSXML2_TLB_Light;

type
  TRTFFont = class(TObject)
    Name: WideString;
    FontName: WideString;
    Charset: Cardinal;
  end;

  TFontList = class(TObjectList)
  private
    function GetFont(const FontName: WideString): TRTFFont; overload;
    function GetFont(Index: Integer): TRTFFont; overload;
    function IndexOf(const FontName: WideString): Integer;
  public
    procedure Add(const Name, FontName: WideString; Charset: Cardinal);
    property Items[const Name: WideString]: TRTFFont read GetFont; default;
  end;

  TRTFToXML = class(TObject)
  private
    FSource: WideString;
    FNChar: Integer;
    FBraceCount: Integer;

    FFontList: TFontList;
    FColors: TStringList;
    FUnicodeLength: Integer;
    FParagraph: IXMLDOMNode;
    FTitle: WideString;
    FStyleChanged: Boolean;

    FSkipStars: Boolean;
    FNextSymbol: Integer;

    FDefaultStyles, FCurrentStyles: TSpanValues;
    function Init: Boolean;
    function HexToInt(const S: WideString): Byte;

    function IsAlphaChar(WC: WideChar): Boolean; inline;
    function IsLowerAlphaChar(WC: WideChar): Boolean; inline;
    function IsUpperAlphaChar(WC: WideChar): Boolean; inline;
    function IsNumeric(WC: WideChar): Boolean; inline;
    function IsControlChar(WC: WideChar): Boolean; inline;
    function GetWideChar(const Value: WideString): WideChar;
    procedure AddBlock(const Parent, Block, Value, Text: WideString;
      IntValue: Integer);
    procedure AddColors(const ColorString: WideString);
    procedure SkipStar;
    function EncodeText(const Text: WideString; Charset: Cardinal): WideString;
    procedure SetNodeText(const Text: WideString; const Node: IXMLDOMNode);
    procedure AddNodeText(const Text: WideString; const Node: IXMLDOMNode);
    procedure AddSpan(const Text: WideString; Unicode: Boolean = False);
    procedure ParseControl(const Control, NumericValue: WideString);
    procedure ParseText(const Control: WideString; var TextValue: WideString);
    procedure SetAttributes(const Node: IXMLDOMNode);
    procedure SetDefaultStyles;
    procedure ParseTitle;
    procedure ParseSymbolFont;
    procedure ProcessFontTable(const BlockName, BlockValue: WideString);
    procedure ProcessGroupRecursevly(const BlockName, BlockValue: WideString);
  protected
  public
    constructor Create;
    destructor Destroy; override;
    function ValidRTF(const Source: WideString): Boolean;
    function Parse(const Source: WideString; 
      var Title: WideString): IXMLDOMDocument;
  end;

implementation

uses
  Windows, SysUtils, Graphics,
  MWKSysUtils, XMLHelp;

{ TRTFToXML }

procedure TRTFToXML.AddBlock(const Parent, Block, Value, Text: WideString;
  IntValue: Integer);
begin
  if Parent = 'fonttbl' then
    FFontList.Add(Block + Value, Copy(Text, 1, Length(Text) - 1), IntValue);

  if Parent = 'colortbl' then
    AddColors(Text);
end;

procedure TRTFToXML.AddColors(const ColorString: WideString);
var
  C, D: Integer;
  SL: TStringList;
  S: WideString;
  Control, NumericValue: WideString;
  R, G, B: WideString;
begin
  SL := TStringList.Create;
  SL.Delimiter := ';';
  SL.DelimitedText := ColorString;
  for C := 0 to SL.Count - 1 do
  begin
    S := SL[C];
    if S = '' then
      FColors.Add('$000000')
    else
    begin
      R := '00';
      G := '00';
      B := '00';
      D := 1;
      while D <= Length(S) do
      case S[D] of
        '\':
        begin
          Inc(D);

          Control := '';
          while IsLowerAlphaChar(S[D]) do
          begin
            Control := Control + S[D];
            Inc(D);
          end;

          if S[D] = '-' then
          begin
            NumericValue := S[D];
            Inc(D);
          end else
            NumericValue := '';

          while IsNumeric(S[D]) do
          begin
            NumericValue := NumericValue + S[D];
            Inc(D);
          end;

          if Control = 'red' then
            R := SysUtils.IntToHex(StrToInt(NumericValue), 2);

          if Control = 'green' then
            G := SysUtils.IntToHex(StrToInt(NumericValue), 2);

          if Control = 'blue' then
            B := SysUtils.IntToHex(StrToInt(NumericValue), 2);
        end;

        else
          Inc(D);
      end;

      FColors.Add('$' + B + G + R);
    end;
  end;
  SL.Free;
end;

procedure TRTFToXML.AddNodeText(const Text: WideString;
  const Node: IXMLDOMNode);
var
  C: Integer;
  S: WideString;
begin
  if HasAttribute(Node, 'sa') then
    S := Node.text +
      DupeString(' ', Node.attributes.getNamedItem('sa').nodeValue) + Text
  else
    S := Node.text + Text;

  C := CountLastSpaces(S);
  if C > 0 then
    SetAttribute(Node, 'sa', C)
  else
    RemoveAttribute(Node, 'sa');

  Node.text := Trim(S);
end;

procedure TRTFToXML.AddSpan(const Text: WideString; Unicode: Boolean = False);
var
  Node: IXMLDOMNode;
  UniText: WideString;
begin
  if Unicode then
    UniText := Text
  else
    UniText := EncodeText(Text, FCurrentStyles.Charset);

  if not FStyleChanged then
  begin
    Node := FParagraph.lastChild;
    if Assigned(Node) and (Node.nodeName = 'span') then
    begin
      AddNodeText(UniText, Node);
      Exit;
    end;
  end;

  Node := AddChildNode(FParagraph, 'span');

  SetNodeText(UniText, Node);

  SetAttributes(Node);
  FStyleChanged := False;
end;

constructor TRTFToXML.Create;
begin
  inherited;
  FColors := TStringList.Create;
  FFontList := TFontList.Create;
  FSkipStars := True;
  FNextSymbol := -1;
end;

destructor TRTFToXML.Destroy;
begin
  FColors.Free;
  FFontList.Free;
  inherited;
end;

function ConvertText(const Text: AnsiString; Charset: Cardinal): WideString;
var
  CharsetInfo: TCharsetInfo;
  L: Integer;
begin
  // Codepage besorgen
  TranslateCharsetInfo(Charset, CharsetInfo, TCI_SRCCHARSET);

  // Puffergre ermitteln
  L := MultiByteToWideChar(CharsetInfo.ciACP, MB_PRECOMPOSED,
    @Text[1], Length(Text), nil, 0);

  // Puffer anlegen
  SetLength(Result, L);

  // Konvertieren
  MultiByteToWideChar(CharsetInfo.ciACP, MB_PRECOMPOSED,
    @Text[1], Length(Text), @Result[1], L);
end;

function TRTFToXML.EncodeText(const Text: WideString;
  Charset: Cardinal): WideString;
var
  C: Integer;
  S: AnsiString;
begin
  SetLength(S, Length(Text));
  for C := 1 to Length(Text) do
    S[C] := Char(Text[C]);

  Result := ConvertText(S, Charset);
end;

function TRTFToXML.GetWideChar(const Value: WideString): WideChar;
begin
  Result := WideChar(StrToInt(Value));
end;

function TRTFToXML.HexToInt(const S: WideString): Byte;

  function GetByte(C: WideChar): Byte;
  begin
    if IsNumeric(C) then
      Result := Ord(C) - Ord('0')
    else
      if (C in [WideChar('A')..WideChar('F')]) then
        Result := Ord(C) - Ord('A') + 10
      else
        if (C in [WideChar('a')..WideChar('f')]) then
          Result := Ord(C) - Ord('a') + 10
        else
          Result := 0;
  end;

begin
  Result := GetByte(S[1]) shl 4 + GetByte(S[2]);
end;

function TRTFToXML.Init: Boolean;
var
  BracesCount: Integer;
  Escaped: Boolean;
  Source: WideString;
  NChar: Integer;
begin
  if Copy(FSource, 1, 5) <> '{\rtf' then
    Result := False
  else
  begin
    Source := '';
    BracesCount := 0;
    Escaped := False;
    NChar := 1;
    while (NChar <= Length(FSource)) and (BracesCount >= 0) do
    begin
      if not (FSource[NChar] in [WideChar(#$0D), WideChar(#$0A)]) then
      begin
        Source := Source + FSource[NChar];
        case FSource[NChar] of
          '{':
            if not Escaped then
              Inc(BracesCount)
            else
              Escaped := False;

          '}':
            if not Escaped then
              Dec(BracesCount)
            else
              Escaped := False;

          '\':
            Escaped := not Escaped;

          else
            Escaped := False;
        end;
      end else
        Escaped := False;
      Inc(NChar);
    end;
    Result := BracesCount = 0;
  end;
  FSource := Source;
end;

function TRTFToXML.IsAlphaChar(WC: WideChar): Boolean;
begin
  Result := IsLowerAlphaChar(WC) or IsUpperAlphaChar(WC);
end;

function TRTFToXML.IsControlChar(WC: WideChar): Boolean;
begin
  Result := WC in [WideChar('{'), WideChar('}'), WideChar('\')];
end;

function TRTFToXML.IsLowerAlphaChar(WC: WideChar): Boolean;
begin
  Result := WC in [WideChar('a')..WideChar('z')];
end;

function TRTFToXML.IsNumeric(WC: WideChar): Boolean;
begin
  Result := WC in [WideChar('0')..WideChar('9')];
end;

function TRTFToXML.IsUpperAlphaChar(WC: WideChar): Boolean;
begin
  Result := WC in [WideChar('A')..WideChar('Z')];
end;

function TRTFToXML.Parse(const Source: WideString;
  var Title: WideString): IXMLDOMDocument;
begin
  FSource := Source;

  if Init then
  begin
    FUnicodeLength := 1;
    FNChar := 1;
    FillChar(FDefaultStyles, SizeOf(FDefaultStyles), 0);
    FDefaultStyles.IsBold := False;
    FDefaultStyles.IsItalic := False;
    FDefaultStyles.IsUnderline := False;
    FDefaultStyles.IsStrikeOut := False;
    FDefaultStyles.IsSpaced := False;
    FDefaultStyles.IsSuper := False;
    FDefaultStyles.IsSub := False;

    FTitle := Title;

    Result := CreateXML(True, 'mwkedit');
    FParagraph := AddChildNode(Result.documentElement, 'p');
    SetDefaultStyles;
    SetAttributes(FParagraph);

    ProcessGroupRecursevly('', '');

    Title := FTitle;
    FParagraph := nil;
  end else
    Result := nil;
end;

procedure TRTFToXML.ParseControl(const Control, NumericValue: WideString);
var
  Value: WideString;
  Num: Integer;
  Node: IXMLDOMNode;
  Font: TRTFFont;
begin
  if (Control = 'line') then
  begin
    Node := AddChildNode(FParagraph, 'br');
    SetAttributes(Node);
    Exit;
  end;

  if (Control = 'par') then
  begin
    FParagraph := AddChildNode(FParagraph.ownerDocument.documentElement, 'p');
    SetAttributes(FParagraph);
    Exit;
  end;

  if Control = '~' then
  begin
    AddSpan(cProtectSpace);
    Exit;
  end;

  if Control = '-' then
  begin
    AddSpan(cHyphen);
    Exit;
  end;

  if Control = 'ldblquote' then
  begin
    AddSpan(cUpperQuote, True);
    Exit;
  end;

  if Control = 'lquote' then
  begin
    AddSpan(cLeftQuote, True);
    Exit;
  end;

  if Control = 'rquote' then
  begin
    AddSpan(cRightQuote, True);
    Exit;
  end;


  if Control = 'tab' then
  begin
    Node := AddChildNode(FParagraph, 'tab');
    Node.text := cHyphen;
    SetAttributes(Node);
    Exit;
  end;

  if Control = 'uc' then
  begin
    if NumericValue <> '' then
      FUnicodeLength := StrToInt(NumericValue);
    Exit;
  end;

  if Control = 'u' then
  begin
    if NumericValue <> '' then
      AddSpan(GetWideChar(NumericValue));
    Exit;
  end;

  if Control = 'plain' then
  begin
    FCurrentStyles.IsBold := False;
    FCurrentStyles.IsItalic := False;
    FCurrentStyles.IsUnderline := False;
    FCurrentStyles.IsStrikeOut := False;
    FCurrentStyles.IsSpaced := False;
    FCurrentStyles.IsSuper := False;
    FCurrentStyles.IsSub := False;
    FStyleChanged := True;
    Exit;
  end;

  if Control = 'fs' then
  begin
    if NumericValue <> '' then
    begin
      FCurrentStyles.FontHeight := StrToInt(NumericValue) div 2;
      FStyleChanged := True;
    end;
    Exit;
  end;

  if Control = 'cf' then
  begin
    if NumericValue <> '' then
      Num := StrToInt(NumericValue)
    else
      Num := 0;

    if (Num >= 0) and (Num < FColors.Count) then
    begin
      Value := FColors[Num];
      FCurrentStyles.Color := StrToInt(Value);
      FStyleChanged := True;
    end;
    Exit;
  end;

  if Control = 'cb' then
  begin
    if NumericValue <> '' then
      Num := StrToInt(NumericValue)
    else
      Num := 0;

    if (Num >= 0) and (Num < FColors.Count) then
    begin
      Value := FColors[Num];
      FCurrentStyles.BackColor := StrToInt(Value);
      FStyleChanged := True;
    end;
    Exit;
  end;

  if Control = 'f' then
  begin
    Font := FFontList[Control + NumericValue];
    if Assigned(Font) then
    begin
      FCurrentStyles.FontName := Font.FontName;
      FCurrentStyles.Charset := Font.Charset;
      FStyleChanged := True;
    end;
    Exit;
  end;

  if Control = 'b' then
  begin
    if ((NumericValue = '') or (StrToInt(NumericValue) > 0)) <> FCurrentStyles.IsBold then
    begin
      FCurrentStyles.IsBold := not FCurrentStyles.IsBold;
      FStyleChanged := True;
    end;
    Exit;
  end;

  if Control = 'i' then
  begin
    if ((NumericValue = '') or (StrToInt(NumericValue) > 0)) <> FCurrentStyles.IsItalic then
    begin
      FCurrentStyles.IsItalic := not FCurrentStyles.IsItalic;
      FStyleChanged := True;
    end;
    Exit;
  end;

  if Control = 'ul' then
  begin
    if ((NumericValue = '') or (StrToInt(NumericValue) > 0)) <> FCurrentStyles.IsUnderline then
    begin
      FCurrentStyles.IsUnderline := not FCurrentStyles.IsUnderline;
      FStyleChanged := True;
    end;
    Exit;
  end;

  if Control = 'strike' then
  begin
    if ((NumericValue = '') or (StrToInt(NumericValue) > 0)) <> FCurrentStyles.IsStrikeOut then
    begin
      FCurrentStyles.IsStrikeOut := not FCurrentStyles.IsStrikeOut;
      FStyleChanged := True;
    end;
    Exit;
  end;

  if (Control = 'super') and (not FCurrentStyles.IsSuper) then
  begin
    FCurrentStyles.IsSuper := True;
    FCurrentStyles.IsSub := False;
    FStyleChanged := True;
    Exit;
  end;

  if (Control = 'sub') and (not FCurrentStyles.IsSub) then
  begin
    FCurrentStyles.IsSub := True;
    FCurrentStyles.IsSuper := False;
    FStyleChanged := True;
    Exit;
  end;

  if (Control = 'nosupersub') and (FCurrentStyles.IsSuper or FCurrentStyles.IsSub) then
  begin
    FCurrentStyles.IsSuper := False;
    FCurrentStyles.IsSub := False;
    FStyleChanged := True;
    Exit;
  end;

  if Control = 'expnd' then
  begin
    if ((NumericValue = '') or (StrToInt(NumericValue) > 0)) <> FCurrentStyles.IsSpaced then
    begin
      FCurrentStyles.IsSpaced := not FCurrentStyles.IsSpaced;
      FStyleChanged := True;
    end;
    Exit;
  end;
end;

procedure TRTFToXML.ParseSymbolFont;
var
  FontName: WideString;
begin
  FontName := '';
  while FSource[FNChar] = ' ' do
    Inc(FNChar);
  if FSource[FNChar] = '"' then
  begin
    Inc(FNChar);

    while FSource[FNChar] <> '"' do
    begin
      FontName := FontName + FSource[FNChar];
      Inc(FNChar);
    end;
    Inc(FNChar);
  end
  else
  begin
    while IsAlphaChar(FSource[FNChar]) do
    begin
      FontName := FontName + FSource[FNChar];
      Inc(FNChar);
    end;
  end;
  if (FontName <> '') then
  begin
    FStyleChanged := True;
    FCurrentStyles.ShadowFontName := FontName;
    AddSpan(Chr(FNextSymbol));
    FStyleChanged := True;
    FCurrentStyles.ShadowFontName := '';
  end;

  FNextSymbol := -1;
  SkipStar;
end;

procedure TRTFToXML.ParseText(const Control: WideString;
  var TextValue: WideString);
begin
  if (Control = 'author') then
  begin
    TextValue := '';
  end;

  if (Control = 'operator') then
  begin
    TextValue := '';
  end;

end;

procedure TRTFToXML.ParseTitle;
var
  TextValue: WideString;
begin
  Inc(FNChar);

  while FNChar <= Length(FSource) do
    case FSource[FNChar] of
      '{':
        ProcessGroupRecursevly('title', '');

      '}':
        Break;

      '\':
      begin
        Inc(FNChar);
        case FSource[FNChar] of
          '''':
          begin
            TextValue := Copy(FSource, FNChar + 1, 2);

            if (FNChar + 5 < Length(FSource)) and (FSource[FNChar + 3] = '\') and
               (FSource[FNChar + 4] = '''') then
            begin
              FTitle := FTitle +
                EncodeText(Chr(HexToInt(TextValue)) + Chr(HexToInt(Copy(FSource, FNChar + 5, 2))), FCurrentStyles.CharSet);
              Inc(FNChar, 4);
            end else
              FTitle := FTitle + WideChar(HexToInt(TextValue));
            Inc(FNChar, 3);
          end;

          '*':
            if FSkipStars then
              SkipStar
            else
               Inc(FNChar);
        end;
      end;

      else
      begin
        FTitle := FTitle + FSource[FNChar];
        Inc(FNChar);
      end;
    end;
end;

procedure TRTFToXML.ProcessFontTable(const BlockName, BlockValue: WideString);
var
  Control, FirstControl: WideString;
  NumericValue, FirstNumeric: WideString;
  TextValue: WideString;
  Charset: Integer;
begin
  Charset := 0;
  Inc(FNChar);
  while FNChar <= Length(FSource) do
    case FSource[FNChar] of
      '{':
        ProcessFontTable(FirstControl, FirstNumeric);

      '}':
      begin
        Inc(FNChar);
        Break;
      end;

      '\':
      begin
        Inc(FNChar);
        case FSource[FNChar] of
          '*':
            if FSkipStars then
              SkipStar
            else
               Inc(FNChar);

          'a'..'z':
          begin
            Control := '';
            while IsAlphaChar(FSource[FNChar]) do
            begin
              Control := Control + FSource[FNChar];
              Inc(FNChar);
            end;

            if FSource[FNChar] = '-' then
            begin
              NumericValue := FSource[FNChar];
              Inc(FNChar);
            end else
              NumericValue := '';

            while IsNumeric(FSource[FNChar]) do
            begin
              NumericValue := NumericValue + FSource[FNChar];
              Inc(FNChar);
            end;

            if FirstControl = '' then
            begin
              FirstControl := Control;
              FirstNumeric := NumericValue;
            end;

            if Control = 'fcharset' then
              TryStrToInt(NumericValue, Charset);

            if FSource[FNChar] = '{' then
              ProcessFontTable(FirstControl, FirstNumeric);

            TextValue := '';
            if not (IsAlphaChar(FSource[FNChar]) or
                    IsControlChar(FSource[FNChar])) then
            begin
              if (FNChar = 1) or (FSource[FNChar - 1] <> '}') then
                Inc(FNChar);
              while not IsControlChar(FSource[FNChar]) do
              begin
                TextValue := TextValue + FSource[FNChar];
                Inc(FNChar);
              end;
            end;

          end;
        end;
      end;

      else
      begin
        TextValue := TextValue + FSource[FNChar];
        Inc(FNChar);
      end;
    end;
                                         
  if (TextValue <> '') and (TextValue[Length(TextValue)] = ';') then
    AddBlock('fonttbl', FirstControl, FirstNumeric, TextValue, Charset);
end;

procedure TRTFToXML.ProcessGroupRecursevly(const BlockName,
  BlockValue: WideString);
var
  Control, NumericValue, TextValue: WideString;
  FirstControl, FirstNumeric: WideString;
begin
  if BlockName = 'fonttbl' then
  begin
    ProcessFontTable(BlockName, BlockValue);
    Exit;
  end;

  SetDefaultStyles;
  Inc(FNChar);

  while FNChar <= Length(FSource) do
    case FSource[FNChar] of
      '{':
        ProcessGroupRecursevly(FirstControl, FirstNumeric);

      '}':
      begin
        Inc(FNChar);
        Break;
      end;

      '\':
      begin
        Inc(FNChar);
        case FSource[FNChar] of
          '\':
          begin
            Inc(FNChar);
            if (FNextSymbol <> -1) and (FSource[FNChar] = 'f') and
               (FSource[FNChar + 1] = ' ') then
            begin
              Inc(FNChar, 2);
              ParseSymbolFont;
              Break;
            end else
              AddSpan('\');
          end;

          '''':
          begin
            TextValue := Copy(FSource, FNChar + 1, 2);

            if (FNChar + 5 < Length(FSource)) and (FSource[FNChar + 3] = '\') and
               (FSource[FNChar + 4] = '''') then
            begin
              Control := Copy(FSource, FNChar + 5, 2);
              AddSpan(Chr(HexToInt(TextValue)) + Chr(HexToInt(Control)));
              Inc(FNChar, 4);
            end else
              AddSpan(WideChar(HexToInt(TextValue)));
            TextValue := '';
            Inc(FNChar, 3);
          end;

          '~', '-':
          begin
            ParseControl(FSource[FNChar], '');
            Inc(FNChar);
          end;

          '*':
            if FSkipStars then
              SkipStar
            else
               Inc(FNChar);

          'a'..'z':
          begin
            Control := '';
            while IsAlphaChar(FSource[FNChar]) do
            begin
              Control := Control + FSource[FNChar];
              Inc(FNChar);
            end;

            if Control = 'title' then
            begin
              ParseTitle;
              Continue;
            end;

            if FSource[FNChar] = '-' then
            begin
              NumericValue := FSource[FNChar];
              Inc(FNChar);
            end else
              NumericValue := '';

            while IsNumeric(FSource[FNChar]) do
            begin
              NumericValue := NumericValue + FSource[FNChar];
              Inc(FNChar);
            end;

            if FirstControl = '' then
            begin
              FirstControl := Control;
              FirstNumeric := NumericValue;
            end;

            if Control = 'pict' then
            begin
              FBraceCount := 1;
              repeat
                Inc(FNChar);
                if FSource[FNChar] = '{' then
                  Inc(FBraceCount);
                if FSource[FNChar] = '}' then
                  Dec(FBraceCount);
              until FBraceCount = 0;
              Continue;
            end;

            if Control = 'colortbl' then
            begin
              TextValue := '';
              while FSource[FNChar] <> '}' do
              begin
                TextValue := TextValue + FSource[FNChar];
                Inc(FNChar);
              end;
              AddBlock(Control, '', '', TextValue, 0);
              TextValue := '';
              Continue;
            end;

            if Control = 'fldinst' then
            begin
              FSkipStars := True;
              TextValue := '';
              NumericValue := '';
              if FSource[FNChar] = ' ' then
              begin
                while (FSource[FNChar] = ' ') do
                  Inc(FNChar);
                while IsAlphaChar(FSource[FNChar]) do
                begin
                  TextValue := TextValue + FSource[FNChar];
                  Inc(FNChar);
                end;

                if (TextValue = 'SYMBOL') and (FSource[FNChar] = ' ') then
                begin
                  while (FSource[FNChar] = ' ') do
                    Inc(FNChar);
                  while IsNumeric(FSource[FNChar]) do
                  begin
                    NumericValue := NumericValue + FSource[FNChar];
                    Inc(FNChar);
                  end;

                  if (NumericValue <> '') then
                    FNextSymbol := StrToInt(NumericValue);
                end;
              end;
            end;

            if FSource[FNChar] = '{' then
            begin
              if FirstControl = 'field' then
              begin
                FSkipStars := False;
                ProcessGroupRecursevly(FirstControl, FirstNumeric);
                FNextSymbol := -1;
                FSkipStars := True;
              end else
                ProcessGroupRecursevly(FirstControl, FirstNumeric);
            end;

            if Control = 'u' then
            begin
              if (FNChar < Length(FSource)) and
                 (FSource[FNChar] = '\') and
                 (FSource[FNChar + 1] = '''') then
                Inc(FNChar, 4)
              else
                Inc(FNChar, FUnicodeLength);
            end;

            TextValue := '';
            if not (IsAlphaChar(FSource[FNChar]) or
                    IsControlChar(FSource[FNChar])) then
            begin
              if (FNChar = 1) or (FSource[FNChar - 1] <> '}') then
                Inc(FNChar);
              while not IsControlChar(FSource[FNChar]) do
              begin
                TextValue := TextValue + FSource[FNChar];
                Inc(FNChar);
              end;
            end;

            ParseControl(Control, NumericValue);
            ParseText(Control, TextValue);

            if Length(TextValue) > 0 then
              if not ((TextValue[Length(TextValue)] = ';') and
                      (FSource[FNChar] = '}')) and
                 (BlockName <> 'info') then
              begin
                AddSpan(TextValue);
                TextValue := '';
              end else
                AddBlock(BlockName, FirstControl, FirstNumeric, TextValue, 0);
          end;

          else
          begin
            AddSpan(FSource[FNChar]);
            Inc(FNChar);
          end;
        end;
      end;

      else
      begin
        AddSpan(FSource[FNChar]);
        Inc(FNChar);
      end;
    end;
end;

procedure TRTFToXML.SetAttributes(const Node: IXMLDOMNode);
var
  Attributes: TStringList;
begin
  if FCurrentStyles.FontName <> '' then
    SetAttribute(Node, 'font', FCurrentStyles.FontName);

  if FCurrentStyles.FontHeight <> 0 then
    SetAttribute(Node, 'size', FCurrentStyles.FontHeight);

  if FCurrentStyles.ShadowFontName <> '' then
    SetAttribute(Node, 'shadowfont', FCurrentStyles.ShadowFontName);

  if FCurrentStyles.Color <> clBlack then
    SetAttribute(Node, 'color', '#' + IntToHex(ColorToRGB(FCurrentStyles.Color), 6));

  if FCurrentStyles.BackColor <> clWhite then
    SetAttribute(Node, 'bg-color', '#' + IntToHex(ColorToRGB(FCurrentStyles.BackColor), 6));

  if FCurrentStyles.Indent <> 0 then
    SetAttribute(Node, 'idt', FCurrentStyles.Indent);

  Attributes := TStringList.Create;
  if FCurrentStyles.IsBold then
    Attributes.Add('bold');

  if FCurrentStyles.IsItalic then
    Attributes.Add('italic');

  if FCurrentStyles.IsUnderline then
    Attributes.Add('underline');

  if FCurrentStyles.IsStrikeOut then
    Attributes.Add('strikeout');

  if FCurrentStyles.IsSuper then
    Attributes.Add('super');

  if FCurrentStyles.IsSub then
    Attributes.Add('sub');

  if FCurrentStyles.IsSpaced then
    Attributes.Add('spaced');

  if Attributes.Count > 0 then
    SetAttribute(Node, 'deco', Attributes.CommaText);

  if FCurrentStyles.IsIndex then
    SetAttribute(Node, 'index', True);

  Attributes.Free;
end;

procedure TRTFToXML.SetDefaultStyles;
begin
  FillChar(FCurrentStyles, SizeOf(FCurrentStyles), 0);
  FCurrentStyles.FontName := FDefaultStyles.FontName;
  FCurrentStyles.Charset:= 0;
  FCurrentStyles.Color := clBlack;
  FCurrentStyles.BackColor := clWhite;
  FStyleChanged := True;
end;

procedure TRTFToXML.SetNodeText(const Text: WideString;
  const Node: IXMLDOMNode);
var
  C: Integer;
begin
  C := CountFirstSpaces(Text);
  if C > 0 then
    SetAttribute(Node, 'sb', C)
  else
    RemoveAttribute(Node, 'sb');

  C := CountLastSpaces(Text);
  if C > 0 then
    SetAttribute(Node, 'sa', C)
  else
    RemoveAttribute(Node, 'sa');

  Node.text := Trim(Text);
end;

procedure TRTFToXML.SkipStar;
var
  BracesOpened: Integer;
  Escaped: Boolean;
begin
  BracesOpened := 1;
  Escaped := False;

  while BracesOpened > 0 do
  begin
    Inc(FNChar);
    case FSource[FNChar] of
      '{':
      if Escaped then
        Escaped := False
      else
        Inc(BracesOpened);

      '}':
      if Escaped then
        Escaped := False
      else
        Dec(BracesOpened);

      '\':
      Escaped := not Escaped;

      else
        Escaped := False;
    end;
  end;
end;

function TRTFToXML.ValidRTF(const Source: WideString): Boolean;
begin
  FSource := Source;

  Result := Init;
end;

{ TFontList }

procedure TFontList.Add(const Name, FontName: WideString; Charset: Cardinal);
var
  RTFFont: TRTFFont;
begin
  RTFFont := GetFont(Name);
  if not Assigned(RTFFont) then
  begin
    RTFFont := TRTFFont.Create;
    RTFFont.Name := Name;
    inherited Add(RTFFont);
  end;

  RTFFont.FontName := FontName;
  RTFFont.Charset := Charset;
end;

function TFontList.GetFont(const FontName: WideString): TRTFFont;
begin
  Result := GetFont(IndexOf(FontName));
end;

function TFontList.GetFont(Index: Integer): TRTFFont;
begin
  if (Index >= 0) and (Index < Count) then
    Result := TRTFFont(inherited Items[Index])
  else
    Result := nil;
end;

function TFontList.IndexOf(const FontName: WideString): Integer;
begin
  for Result := 0 to Count - 1 do
    if GetFont(Result).Name = FontName then
      Exit;
  Result := -1;
end;

end.
